@@ -15,7 +15,7 @@ class AudioSourceIO : public QIODevice
 {
     Q_OBJECT
 public:
-    AudioSourceIO(QFFmpegAudioInput *audioInput) : QIODevice(), m_input(audioInput)
+    AudioSourceIO(QFFmpegAudioInput *audioInput) : m_input(audioInput)
     {
         m_muted = m_input->muted;
         m_volume = m_input->volume;
@@ -23,7 +23,13 @@ public:
         open(QIODevice::WriteOnly);
     }
 
-    ~AudioSourceIO() override = default;
+    ~AudioSourceIO() override
+    {
+        // QAudioSource may invoke QIODevice::writeData in the destructor.
+        // Let's reset the audio source to get around the case.
+        if (m_audioSource)
+            m_audioSource->reset();
+    }
 
     void setDevice(const QAudioDevice &device)
     {
@@ -33,10 +39,11 @@ public:
         m_device = device;
         QMetaObject::invokeMethod(this, "updateSource");
     }
-    void setFrameSize(int frameSize)
+    void setBufferSize(int bufferSize)
     {
-        m_bufferSize.storeRelease(frameSize > 0 ? m_format.bytesForFrames(frameSize)
-                                                : DefaultAudioInputBufferSize);
+        m_bufferSize.storeRelease((bufferSize > 0 && m_format.isValid())
+                                          ? m_format.bytesForFrames(bufferSize)
+                                          : DefaultAudioInputBufferSize);
     }
     void setRunning(bool r) {
         QMutexLocker locker(&m_mutex);
@@ -66,15 +73,28 @@ protected:
     }
     qint64 writeData(const char *data, qint64 len) override
     {
+        Q_ASSERT(m_audioSource);
+
         int l = len;
         while (len > 0) {
             const auto bufferSize = m_bufferSize.loadAcquire();
+
+            while (m_pcm.size() > bufferSize) {
+                // bufferSize has been reduced. Send data until m_pcm
+                // can hold more data.
+                sendBuffer(m_pcm.first(bufferSize));
+                m_pcm.remove(0, bufferSize);
+            }
+
+            // Size of m_pcm is always <= bufferSize
             int toAppend = qMin(len, bufferSize - m_pcm.size());
             m_pcm.append(data, toAppend);
             data += toAppend;
             len -= toAppend;
-            if (m_pcm.size() == bufferSize)
-                sendBuffer();
+            if (m_pcm.size() == bufferSize) {
+                sendBuffer(m_pcm);
+                m_pcm.clear();
+            }
         }
 
         return l;
@@ -84,41 +104,40 @@ private Q_SLOTS:
     void updateSource() {
         QMutexLocker locker(&m_mutex);
         m_format = m_device.preferredFormat();
-        if (std::exchange(m_src, nullptr))
+        if (std::exchange(m_audioSource, nullptr))
             m_pcm.clear();
 
-        m_src = std::make_unique<QAudioSource>(m_device, m_format);
+        m_audioSource = std::make_unique<QAudioSource>(m_device, m_format);
         updateVolume();
         if (m_running)
-            m_src->start(this);
+            m_audioSource->start(this);
     }
     void updateVolume()
     {
-        if (m_src)
-            m_src->setVolume(m_muted ? 0. : m_volume);
+        if (m_audioSource)
+            m_audioSource->setVolume(m_muted ? 0. : m_volume);
     }
     void updateRunning()
     {
         QMutexLocker locker(&m_mutex);
         if (m_running) {
-            if (!m_src)
+            if (!m_audioSource)
                 updateSource();
-            m_src->start(this);
+            m_audioSource->start(this);
         } else {
-            m_src->stop();
+            m_audioSource->stop();
         }
     }
 
 private:
 
-    void sendBuffer()
+    void sendBuffer(const QByteArray &pcmData)
     {
-        QAudioFormat fmt = m_src->format();
+        QAudioFormat fmt = m_audioSource->format();
         qint64 time = fmt.durationForBytes(m_processed);
-        QAudioBuffer buffer(m_pcm, fmt, time);
+        QAudioBuffer buffer(pcmData, fmt, time);
         emit m_input->newAudioBuffer(buffer);
-        m_processed += m_pcm.size();
-        m_pcm.clear();
+        m_processed += pcmData.size();
     }
 
     QMutex m_mutex;
@@ -128,60 +147,63 @@ private:
     bool m_running = false;
 
     QFFmpegAudioInput *m_input = nullptr;
-    std::unique_ptr<QAudioSource> m_src;
+    std::unique_ptr<QAudioSource> m_audioSource;
     QAudioFormat m_format;
     QAtomicInt m_bufferSize = DefaultAudioInputBufferSize;
     qint64 m_processed = 0;
     QByteArray m_pcm;
 };
 
-}
+} // namespace QFFmpeg
 
 QFFmpegAudioInput::QFFmpegAudioInput(QAudioInput *qq)
     : QPlatformAudioInput(qq)
 {
     qRegisterMetaType<QAudioBuffer>();
 
-    inputThread = std::make_unique<QThread>();
-    audioIO = std::make_unique<QFFmpeg::AudioSourceIO>(this);
-    audioIO->moveToThread(inputThread.get());
-    inputThread->start();
+    m_inputThread = std::make_unique<QThread>();
+    m_audioIO = new QFFmpeg::AudioSourceIO(this);
+    m_audioIO->moveToThread(m_inputThread.get());
+    m_inputThread->start();
 }
 
 QFFmpegAudioInput::~QFFmpegAudioInput()
 {
-    inputThread->exit();
-    inputThread->wait();
+    // Ensure that COM is uninitialized by nested QWindowsResampler
+    // on the same thread that initialized it.
+    m_audioIO->deleteLater();
+    m_inputThread->exit();
+    m_inputThread->wait();
 }
 
 void QFFmpegAudioInput::setAudioDevice(const QAudioDevice &device)
 {
-    audioIO->setDevice(device);
+    m_audioIO->setDevice(device);
 }
 
 void QFFmpegAudioInput::setMuted(bool muted)
 {
-    audioIO->setMuted(muted);
+    m_audioIO->setMuted(muted);
 }
 
 void QFFmpegAudioInput::setVolume(float volume)
 {
-    audioIO->setVolume(volume);
+    m_audioIO->setVolume(volume);
 }
 
-void QFFmpegAudioInput::setFrameSize(int s)
+void QFFmpegAudioInput::setBufferSize(int bufferSize)
 {
-    audioIO->setFrameSize(s);
+    m_audioIO->setBufferSize(bufferSize);
 }
 
 void QFFmpegAudioInput::setRunning(bool b)
 {
-    audioIO->setRunning(b);
+    m_audioIO->setRunning(b);
 }
 
 int QFFmpegAudioInput::bufferSize() const
 {
-    return audioIO->bufferSize();
+    return m_audioIO->bufferSize();
 }
 
 QT_END_NAMESPACE
